/*
 *        nProbe - a Netflow v5/v9/IPFIX probe for IPv4/v6
 *
 *       Copyright (C) 2002-13 Luca Deri <deri@ntop.org>
 *
 *                     http://www.ntop.org/
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License along
 *  with this program; if not, write to the Free Software Foundation, Inc.,
 *  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

#include "nprobe.h"

static u_int8_t traceLRU = 0;

/* ************************************ */

static void incrementQueueStats(void) {
#ifdef HAVE_REDIS
  readOnlyGlobals.redis.queuedSetDeleteCommands++;
  
  if(readOnlyGlobals.redis.queuedSetDeleteCommands > readOnlyGlobals.redis.maxQueuedSetDeleteCommands)
    readOnlyGlobals.redis.maxQueuedSetDeleteCommands = readOnlyGlobals.redis.queuedSetDeleteCommands;
#endif
}

/* ************************************ */

void setCacheKeyValueString(const char *prefix, const char *key, const char *value) {
#ifdef HAVE_REDIS
  if(readOnlyGlobals.redis.context) {
    pthread_rwlock_wrlock(&readOnlyGlobals.redis.lock_set_delete);
    if(unlikely(readOnlyGlobals.enable_debug)) traceEvent(TRACE_NORMAL, "[Redis] SET %s%s %s", prefix, key, value);
    redisAppendCommand(readOnlyGlobals.redis.set_delete_context, "SET %s%s %s", prefix, key, value);
    incrementQueueStats();
    pthread_rwlock_unlock(&readOnlyGlobals.redis.lock_set_delete);
  }
#endif
}

/* ************************************************* */

void setCacheKeyValueNumber(const char *prefix, const char *key, u_int64_t value) {
#ifdef HAVE_REDIS
  if(readOnlyGlobals.redis.context) {
    pthread_rwlock_wrlock(&readOnlyGlobals.redis.lock_set_delete);
    if(unlikely(readOnlyGlobals.enable_debug)) traceEvent(TRACE_NORMAL, "[Redis] SET %s%s %llu", prefix, key, value);
    redisAppendCommand(readOnlyGlobals.redis.set_delete_context, "SET %s%s %llu", prefix, key, value);
    incrementQueueStats();
    pthread_rwlock_unlock(&readOnlyGlobals.redis.lock_set_delete);
  }
#endif
}

/* ************************************************* */

void incrCacheKeyValueNumber(const char *prefix, const char *key, u_int64_t value) {
#ifdef HAVE_REDIS
  if(readOnlyGlobals.redis.context) {
    pthread_rwlock_wrlock(&readOnlyGlobals.redis.lock_set_delete);
    if(unlikely(readOnlyGlobals.enable_debug)) traceEvent(TRACE_NORMAL, "[Redis] INCRBY %s%s %llu", prefix, key, value);
    redisAppendCommand(readOnlyGlobals.redis.set_delete_context, "INCRBY %s%s %llu", prefix, key, value);
    incrementQueueStats();
    pthread_rwlock_unlock(&readOnlyGlobals.redis.lock_set_delete);
  }
#endif
}

/* ************************************************* */

void incrHashCacheKeyValueNumber(const char *element, const char *key, u_int64_t value) {
#ifdef HAVE_REDIS
  if(readOnlyGlobals.redis.context) {
    pthread_rwlock_wrlock(&readOnlyGlobals.redis.lock_set_delete);
    if(unlikely(readOnlyGlobals.enable_debug)) traceEvent(TRACE_NORMAL, "[Redis] HINCRBY %s %s %llu", element, key, value);
    redisAppendCommand(readOnlyGlobals.redis.set_delete_context, "HINCRBY %s %s %llu", element, key, value);
    incrementQueueStats();
    pthread_rwlock_unlock(&readOnlyGlobals.redis.lock_set_delete);
  }
#endif
}

/* ************************************************* */

void expireCacheKey(const char *prefix, const char *key, u_int32_t duration_sec) {
#ifdef HAVE_REDIS
  if(readOnlyGlobals.redis.context) {
    pthread_rwlock_wrlock(&readOnlyGlobals.redis.lock_set_delete);
    if(unlikely(readOnlyGlobals.enable_debug)) traceEvent(TRACE_NORMAL, "[Redis] EXPIRE %s%s %u", prefix, key, duration_sec);
    redisAppendCommand(readOnlyGlobals.redis.set_delete_context, "EXPIRE %s%s %u", prefix, key, duration_sec);
    incrementQueueStats();
    pthread_rwlock_unlock(&readOnlyGlobals.redis.lock_set_delete);
  }
#endif
}

/* ************************************ */

void setCacheHashKeyValueString(const char *element, const char *key, const char *value) {
#ifdef HAVE_REDIS
  if(readOnlyGlobals.redis.context) {
    pthread_rwlock_wrlock(&readOnlyGlobals.redis.lock_set_delete);
    if(unlikely(readOnlyGlobals.enable_debug)) traceEvent(TRACE_NORMAL, "[Redis] HSET %s %s %s", element, key, value);
    redisAppendCommand(readOnlyGlobals.redis.set_delete_context, "HSET %s %s %s", element, key, value);
    incrementQueueStats();
    pthread_rwlock_unlock(&readOnlyGlobals.redis.lock_set_delete);
  }
#endif
}

/* ************************************************* */

void setCacheHashKeyValueNumber(const char *element, const char *key, u_int64_t value) {
#ifdef HAVE_REDIS
  if(readOnlyGlobals.redis.context) {
    pthread_rwlock_wrlock(&readOnlyGlobals.redis.lock_set_delete);
    if(unlikely(readOnlyGlobals.enable_debug)) traceEvent(TRACE_NORMAL, "[Redis] HSET %s %s %llu", element, key, value);
    redisAppendCommand(readOnlyGlobals.redis.set_delete_context, "HSET %s %s %llu", element, key, value);
    incrementQueueStats();
    pthread_rwlock_unlock(&readOnlyGlobals.redis.lock_set_delete);
  }
#endif
}

/* ************************************************* */

void incrCacheHashKeyValueNumber(const char *element,
				 const char *key, 
				 const u_int64_t value) {
#ifdef HAVE_REDIS
  if(readOnlyGlobals.redis.context && (value > 0)) {
    pthread_rwlock_wrlock(&readOnlyGlobals.redis.lock_set_delete);
    if(unlikely(readOnlyGlobals.enable_debug)) traceEvent(TRACE_NORMAL, "[Redis] HINCRBY %s %s %llu", element, key, value);
    redisAppendCommand(readOnlyGlobals.redis.set_delete_context, "HINCRBY %s %s %llu", element, key, value);
    incrementQueueStats();
    pthread_rwlock_unlock(&readOnlyGlobals.redis.lock_set_delete);
  }
#endif
}

/* ************************************************* */

void zIncrCacheHashKeyValueNumber(const char *set_name,
				  const char *key, 
				  const u_int64_t value) {
#ifdef HAVE_REDIS
  if(readOnlyGlobals.redis.context && (value > 0)) {
    pthread_rwlock_wrlock(&readOnlyGlobals.redis.lock_set_delete);
    if(unlikely(readOnlyGlobals.enable_debug)) traceEvent(TRACE_NORMAL, "[Redis] ZINCRBY %s %llu %s", set_name, value, key);
    redisAppendCommand(readOnlyGlobals.redis.set_delete_context, "ZINCRBY %s %llu %s", set_name, value, key);
    readOnlyGlobals.redis.queuedSetDeleteCommands++;
    pthread_rwlock_unlock(&readOnlyGlobals.redis.lock_set_delete);
  }
#endif
}

/* ************************************************* */

void setCacheNumKeyMixedValueQuad(const char *prefix,
                                  const u_int32_t key0, const char* value0,
                                  const u_int32_t key1, const char* value1,
                                  const u_int32_t key2, const u_int32_t value2,
                                  const u_int32_t key3, const u_int32_t value3) {
#ifdef HAVE_REDIS
  if(readOnlyGlobals.redis.context) {
    pthread_rwlock_wrlock(&readOnlyGlobals.redis.lock_set_delete);
    if(unlikely(readOnlyGlobals.enable_debug))
      traceEvent(TRACE_NORMAL, "[Redis] MSET %s%u %s %s%u %s %s%u %u %s%u %u",
                 prefix, key0, value0, prefix, key1, value1,
                 prefix, key2, value2, prefix, key3, value3);
    redisAppendCommand(readOnlyGlobals.redis.set_delete_context,
                       "MSET %s%u %s %s%u %s %s%u %u %s%u %u",
                       prefix, key0, value0, prefix, key1, value1,
                       prefix, key2, value2, prefix, key3, value3);
    incrementQueueStats();
    pthread_rwlock_unlock(&readOnlyGlobals.redis.lock_set_delete);
  }
#endif
}

/* ************************************************* */

void setCacheHashNumKeyMixedValueQuad(const char *master_key,
				      const u_int32_t key0, const char* value0,
				      const u_int32_t key1, const char* value1,
				      const u_int32_t key2, const u_int32_t value2,
				      const u_int32_t key3, const u_int32_t value3) {
#ifdef HAVE_REDIS
  if(readOnlyGlobals.redis.context) {
    pthread_rwlock_wrlock(&readOnlyGlobals.redis.lock_set_delete);
    if(unlikely(readOnlyGlobals.enable_debug))
      traceEvent(TRACE_NORMAL, "[Redis] HMSET %s %u %s %u %s %u %u %u %u",
		 master_key,
		 key0, value0,
		 key1, value1,
		 key2, value2, 
		 key3, value3);
    redisAppendCommand(readOnlyGlobals.redis.set_delete_context,
		       "HMSET %s %u %s %u %s %u %u %u %u",
		       master_key,
		       key0, value0,
		       key1, value1,
		       key2, value2,
		       key3, value3);

    incrementQueueStats();
    pthread_rwlock_unlock(&readOnlyGlobals.redis.lock_set_delete);
  }
#endif
}

/* ************************************ */

char* getCacheDataNumKey(const char *prefix, const u_int32_t key) {
#ifdef HAVE_REDIS
  char *ret = NULL;
  redisReply *reply;

  if(readOnlyGlobals.redis.context) {
    pthread_rwlock_wrlock(&readOnlyGlobals.redis.lock_get);
    if(unlikely(readOnlyGlobals.enable_debug)) traceEvent(TRACE_NORMAL, "[Redis] GET %s%u", prefix, key);
    reply = redisCommand(readOnlyGlobals.redis.context, "GET %s%u", prefix, key);
    pthread_rwlock_unlock(&readOnlyGlobals.redis.lock_get);

    if(reply) {
      if(reply->str) {
	ret = strdup(reply->str);
	if(unlikely(readOnlyGlobals.enable_debug))
	  traceEvent(TRACE_NORMAL, "[Redis] %s(%u)=%s", __FUNCTION__, key, ret);
      }
      freeReplyObject(reply);
    }
  }

  return(ret);
#else
  return(NULL);
#endif
}

/* ************************************ */

char* getCacheDataStrKey(const char *prefix, const char *key) {
#ifdef HAVE_REDIS
  char *ret = NULL;
  redisReply *reply;

  if(readOnlyGlobals.redis.context) {
    pthread_rwlock_wrlock(&readOnlyGlobals.redis.lock_get);
    if(unlikely(readOnlyGlobals.enable_debug)) traceEvent(TRACE_NORMAL, "[Redis] GET %s%s", prefix, key);
    reply = redisCommand(readOnlyGlobals.redis.context, "GET %s%s", prefix, key);
    pthread_rwlock_unlock(&readOnlyGlobals.redis.lock_get);

    if(reply) {
      if(reply->str) {
	ret = strdup(reply->str);
	traceEvent(TRACE_INFO, "[Redis] %s(%u)=%s", __FUNCTION__, key, ret);
      }
      freeReplyObject(reply);
    }
  }

  return(ret);
#else
  return(NULL);
#endif
}

/* ************************************ */

int deleteCacheStrKey(const char *prefix, const char *key, const u_int32_t delete_delay_sec) {
#ifdef HAVE_REDIS
  if(readOnlyGlobals.redis.context) {
    if(unlikely(readOnlyGlobals.enable_debug)) traceEvent(TRACE_NORMAL, "[Redis] EXPIRE %s%s %d", prefix, key, delete_delay_sec);
    pthread_rwlock_wrlock(&readOnlyGlobals.redis.lock_set_delete);
    if(delete_delay_sec > 0)
      redisAppendCommand(readOnlyGlobals.redis.set_delete_context, "EXPIRE %s%s %d", prefix, key, delete_delay_sec);
    else
      redisAppendCommand(readOnlyGlobals.redis.set_delete_context, "DEL %s%s", prefix, key);

    incrementQueueStats();
    pthread_rwlock_unlock(&readOnlyGlobals.redis.lock_set_delete);
  }
#endif

  return(0);
}

/* ************************************ */

int deleteCacheNumKey(const char *prefix, const u_int32_t key, const u_int32_t delete_delay_sec) {
#ifdef HAVE_REDIS
  if(readOnlyGlobals.redis.context) {
    pthread_rwlock_wrlock(&readOnlyGlobals.redis.lock_set_delete);
    if(unlikely(readOnlyGlobals.enable_debug)) traceEvent(TRACE_NORMAL, "[Redis] EXPIRE %s%u %d", prefix, key, delete_delay_sec);
    if(delete_delay_sec > 0)
      redisAppendCommand(readOnlyGlobals.redis.set_delete_context, "EXPIRE %s%u %d", prefix, key, delete_delay_sec);
    else
      redisAppendCommand(readOnlyGlobals.redis.set_delete_context, "DEL %s%u", prefix, key);

    incrementQueueStats();
    pthread_rwlock_unlock(&readOnlyGlobals.redis.lock_set_delete);
  }
#endif

  return(0);
}

/* ************************************ */

/* Redis does not accept expire time */
int deleteCacheNumKeyTwin(const char *prefix, const u_int32_t key1, const u_int32_t key2) {
#ifdef HAVE_REDIS
  if(readOnlyGlobals.redis.context) {
    pthread_rwlock_wrlock(&readOnlyGlobals.redis.lock_set_delete);
    if(unlikely(readOnlyGlobals.enable_debug)) traceEvent(TRACE_NORMAL, "[Redis] DEL %s%u %s%u", prefix, key1, prefix, key2);
    redisAppendCommand(readOnlyGlobals.redis.set_delete_context, "DEL %s%u %s%u", prefix, key1, prefix, key2);

    incrementQueueStats();
    pthread_rwlock_unlock(&readOnlyGlobals.redis.lock_set_delete);
  }
#endif

  return(0);
}

/* ************************************ */

/* Redis does not accept expire time */
int deleteCacheStrKeyTwin(const char *prefix, const char *key1, const char *key2) {
#ifdef HAVE_REDIS
  if(readOnlyGlobals.redis.context) {
    pthread_rwlock_wrlock(&readOnlyGlobals.redis.lock_set_delete);
    if(unlikely(readOnlyGlobals.enable_debug)) traceEvent(TRACE_NORMAL, "[Redis] DEL %s%s %s%s", prefix, key1, prefix, key2);
    redisAppendCommand(readOnlyGlobals.redis.set_delete_context, "DEL %s%s %s%s", prefix, key1, prefix, key2);

    incrementQueueStats();
    pthread_rwlock_unlock(&readOnlyGlobals.redis.lock_set_delete);
  }
#endif

  return(0);
}

/* ************************************ */

char* getHashCacheDataStrKey(const char *prefix, const char *element, const char *key) {
#ifdef HAVE_REDIS
  char *ret = NULL;
  redisReply *reply;

  if(readOnlyGlobals.redis.context) {
    pthread_rwlock_wrlock(&readOnlyGlobals.redis.lock_get);
    if(unlikely(readOnlyGlobals.enable_debug)) traceEvent(TRACE_NORMAL, "[Redis] HGET %s%s %s", prefix, element, key);
    reply = redisCommand(readOnlyGlobals.redis.context, "HGET %s%s %s", prefix, element, key);
    pthread_rwlock_unlock(&readOnlyGlobals.redis.lock_get);

    if(reply) {
      if(reply->str) {
	ret = strdup(reply->str);
	traceEvent(TRACE_NORMAL, "[Redis] %s(%u)=%s", __FUNCTION__, key, ret);
      }
      freeReplyObject(reply);
    }
  }

  return(ret);
#else
  return(NULL);
#endif
}

/* ************************************ */

#ifdef HAVE_REDIS

static void processQueuedRedisCommand(void) {
  if(readOnlyGlobals.redis.queuedSetDeleteCommands > 0) {
    redisReply *reply;

    pthread_rwlock_wrlock(&readOnlyGlobals.redis.lock_set_delete);
    redisGetReply(readOnlyGlobals.redis.set_delete_context, (void**)&reply);
    freeReplyObject(reply);
    readOnlyGlobals.redis.queuedSetDeleteCommands--;

#if 0
    if(unlikely(readOnlyGlobals.enable_debug))
      traceEvent(TRACE_NORMAL, "[Redis] %s(): got reply [queue=%d]",
		 __FUNCTION__, readOnlyGlobals.redis.queuedSetDeleteCommands);
#endif

    pthread_rwlock_unlock(&readOnlyGlobals.redis.lock_set_delete);
  }
}
#endif

/* ************************************ */

#define MAX_NUM_ARGUMENTS   32
#define MAX_ARGUMENT_LEN   256

int readRedisLine(int sock, char *line) {
  int pos = 0, rc;

  while((rc = recv(sock, &line[pos++], 1, 0)) > 0) {
    if((pos > 2) 
       && (line[pos-2] == '\r')
       && (line[pos-1] == '\n')) {
      line[pos++] = '\0';
      return(pos);
    }
  }

  return(0);
}

int readRedisCommand(int sock, char request[MAX_NUM_ARGUMENTS][MAX_ARGUMENT_LEN]) {
  /*
   *1
   $4
   PING
   
   *<number of arguments> CR LF
   $<number of bytes of argument 1> CR LF
   <argument data> CR LF
   ...
   $<number of bytes of argument N> CR LF
   <argument data> CR LF
  */
  
  
  readRedisLine(sock, request[0]);
  request[1][0] = '\0';

  return(0);
}

/* ************************************ */

void handleClient(int sock) {
  char request[MAX_NUM_ARGUMENTS][MAX_ARGUMENT_LEN], *rsp;
  
  while(readRedisCommand(sock, request) >= 0) {
    int i;

    for(i=0; request[i][0] != '\0'; i++)
      traceEvent(TRACE_NORMAL, "%s", request[i]);

    rsp = "+OK\r\n";
    send(sock, rsp, strlen(rsp), 0);
  }
}

/* ************************************ */

#ifdef HAVE_REDIS
static void* redisLocalServerLoop(void* notUsed) {
  traceEvent(TRACE_NORMAL, "[Redis Server] %s() started", __FUNCTION__);

  readOnlyGlobals.redis.local_server_running = 0;

  while(!readWriteGlobals->shutdownInProgress) {
    fd_set mask;

    FD_ZERO(&mask);
    FD_SET(readOnlyGlobals.redis.local_server_socket, &mask);

    if(select(readOnlyGlobals.redis.local_server_socket+1, &mask, 0, 0, NULL) > 0) {
      struct sockaddr_in from;
      int rx_sock, from_len, rc;
      char aChar;

      rx_sock = accept(readOnlyGlobals.redis.local_server_socket, (struct sockaddr*)&from, (socklen_t*)&from_len);
      if((rx_sock < 0) || (errno != 0)) {
	traceEvent(TRACE_ERROR, "Unable to accept connection [%s/%d]", strerror(errno), errno);	
      }

      traceEvent(TRACE_NORMAL, "[Redis Server] New client connected [socket %d]", rx_sock);
      handleClient(rx_sock);
      close(rx_sock);
    }
  }

  readOnlyGlobals.redis.local_server_running = 1;
}
#endif

/* ************************************ */

#ifdef HAVE_REDIS
static void* redisAsyncLoop(void* notUsed) {
  traceEvent(TRACE_NORMAL, "[Redis] %s() started", __FUNCTION__);

  readOnlyGlobals.redis.queue_thread_running = 1;

  while(!readWriteGlobals->shutdownInProgress) {
    if(readOnlyGlobals.redis.queuedSetDeleteCommands > 0) {
      processQueuedRedisCommand();
    } else {
      usleep(10000); /* 10 ms */
    }
  }

  /* Flush pending commands */
  while(readOnlyGlobals.redis.queuedSetDeleteCommands > 0)
    processQueuedRedisCommand();

  readOnlyGlobals.redis.queue_thread_running = 0;

  traceEvent(TRACE_INFO, "[Redis] %s() completed [queue=%d]",
	     __FUNCTION__, readOnlyGlobals.redis.queuedSetDeleteCommands);
  return(NULL);
}
#endif

/* ************************************ */

int createLocalCacheServer() {
#ifdef HAVE_REDIS
  int sockopt = 1, rc;
  struct sockaddr_in sockIn;
  struct addrinfo hints, *ai = NULL;

  if(readOnlyGlobals.redis.local_ucloud_port == 0)
    return(0);

  errno = 0;
  readOnlyGlobals.redis.local_server_socket = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
  
  if((readOnlyGlobals.redis.local_server_socket < 0) || (errno != 0)) {
    traceEvent(TRACE_ERROR, "Unable to create server socket [%s/%d]", strerror(errno), errno);
    exit(-1);
  }

  errno = 0;
  rc = setsockopt(readOnlyGlobals.redis.local_server_socket, SOL_SOCKET, SO_REUSEADDR, (char *)&sockopt, sizeof(sockopt));

  memset(&sockIn, 0, sizeof(sockIn));
  sockIn.sin_family = AF_INET, sockIn.sin_port = (int)htons(readOnlyGlobals.redis.local_ucloud_port), errno = 0;
  rc = bind(readOnlyGlobals.redis.local_server_socket, (struct sockaddr *)&sockIn, sizeof(sockIn));
  if((readOnlyGlobals.redis.local_server_socket < 0) || (errno != 0)) {
    traceEvent(TRACE_ERROR, "Unable to bind to the specified port [%s/%d]", strerror(errno), errno);
    exit(-1);
  }

  errno = 0;
  rc = listen(readOnlyGlobals.redis.local_server_socket, 1 /* 1 client max */);

  pthread_create(&readOnlyGlobals.redis.local_server_loop, NULL, redisLocalServerLoop, NULL);
#endif
  return(0);
}

/* ************************************ */

int connectToRemoteCache(void) {
#ifdef HAVE_REDIS
  struct timeval timeout = { 1, 500000 }; // 1.5 seconds

  if(unlikely(readOnlyGlobals.enable_debug))
    traceEvent(TRACE_NORMAL, "[Redis] %s(%s:%u)", __FUNCTION__, 
	       readOnlyGlobals.redis.remote_redis_host, readOnlyGlobals.redis.remote_redis_port);

  /* Synchronous */
  readOnlyGlobals.redis.context = redisConnectWithTimeout(readOnlyGlobals.redis.remote_redis_host, 
							  readOnlyGlobals.redis.remote_redis_port, timeout);
  if(readOnlyGlobals.redis.context->err) {
    traceEvent(TRACE_ERROR, "Redis Connection error: %s",
	       readOnlyGlobals.redis.context->errstr);
    exit(-1);
  }

  /* Asynchronous */
  readOnlyGlobals.redis.set_delete_context = redisConnectWithTimeout(readOnlyGlobals.redis.remote_redis_host, 
								     readOnlyGlobals.redis.remote_redis_port, timeout);
  if(readOnlyGlobals.redis.set_delete_context->err) {
    traceEvent(TRACE_ERROR, "Redis Connection error: %s",
	       readOnlyGlobals.redis.set_delete_context->errstr);
    exit(-1);
  }

  pthread_rwlock_init(&readOnlyGlobals.redis.lock_get, NULL);
  pthread_rwlock_init(&readOnlyGlobals.redis.lock_set_delete, NULL);

  pthread_create(&readOnlyGlobals.redis.reply_loop, NULL, redisAsyncLoop, NULL);

  createLocalCacheServer();
#endif

  return(0);
}

/* ************************************ */

void disconnectFromRemoteCache(void) {
#ifdef HAVE_REDIS
  if(unlikely(readOnlyGlobals.enable_debug))
    traceEvent(TRACE_NORMAL, "[Redis] %s()", __FUNCTION__);

  while(readOnlyGlobals.redis.queuedSetDeleteCommands > 0) {
    if(!readOnlyGlobals.redis.queue_thread_running)
      processQueuedRedisCommand();
    else
      sleep(1);
  }

  if(readOnlyGlobals.redis.context)
    redisFree(readOnlyGlobals.redis.context);

  if(readOnlyGlobals.redis.set_delete_context)
    redisFree(readOnlyGlobals.redis.set_delete_context);

  pthread_rwlock_destroy(&readOnlyGlobals.redis.lock_get);
  pthread_rwlock_destroy(&readOnlyGlobals.redis.lock_set_delete);

  if(readOnlyGlobals.redis.local_server_socket > 0)
    close(readOnlyGlobals.redis.local_server_socket);
#endif
}

/* ************************************ */

void dumpCacheStats() {
#ifdef HAVE_REDIS
  traceEvent(TRACE_NORMAL, "Cache [queuedSetDelete: actual %u/max %u]",
	     readOnlyGlobals.redis.queuedSetDeleteCommands,
	     readOnlyGlobals.redis.maxQueuedSetDeleteCommands);
#endif
}

/* ************************************ */

static void validate_lru_unit_len(struct LruCacheUnit *cache_unit) {
  struct LruCacheNode *tail = cache_unit->list_tail;
  u_int len = 0;

  // traceEvent(TRACE_NORMAL, "%s()", __FUNCTION__);

  while(tail != NULL) {
    len++;
    
    if(tail == tail->lru_list.prev) {
      traceEvent(TRACE_ERROR, "Internal error (loop)");
      break;
    }

    tail = tail->lru_list.prev;
  }

  if(len != cache_unit->current_size)
    traceEvent(TRACE_ERROR, "Internal error [expected len: %u][actual len: %u]",
	       cache_unit->current_size, len);
}

/* ************************************ */

static void validate_lru_len(struct LruCache *cache) {
  int i;

  for(i=0; i<NUM_LRU_CACHE_UNITS; i++)
    validate_lru_unit_len(&cache->units[i]);
}

/* ************************************ */

void test_lru_cache(struct LruCache *cache) {
  u_int64_t key = 1054;
  u_int32_t rc, val = 34, val1 = 689327, n;
  traceEvent(TRACE_NORMAL, "Testing LRU");

  rc = add_to_lru_cache(cache, key, val); if(rc != 0) traceEvent(TRACE_ERROR, "add_to_lru_cache returned %d", rc);
  validate_lru_len(cache);

  rc = find_lru_cache(cache, key); if(rc != val) traceEvent(TRACE_ERROR, "Search found: %u", rc);
  validate_lru_len(cache);

  rc = add_to_lru_cache(cache, key, val1); if(rc != 0) traceEvent(TRACE_ERROR, "add_to_lru_cache returned %d", rc);
  validate_lru_len(cache);

  rc = find_lru_cache(cache, key); if(rc != val1) traceEvent(TRACE_ERROR, "Search found: %u", rc);
  validate_lru_len(cache);

  rc = find_lru_cache(cache, key); if(rc != val1) traceEvent(TRACE_ERROR, "Search found: %u", rc);
  validate_lru_len(cache);
  /* ****************** */
  for(n=cache->units[0].max_lru_size*2; n>0; n--) {
    rc = add_to_lru_cache(cache, key, val1); if(rc != 0) traceEvent(TRACE_ERROR, "add_to_lru_cache returned %d", rc);
    validate_lru_len(cache);
  }

  /* ****************** */
  for(n=cache->units[0].max_lru_size*2; n>0; n--) {

    rc = add_to_lru_cache(cache, n, n); if(rc != 0) traceEvent(TRACE_ERROR, "add_to_lru_cache returned %d", rc);
    validate_lru_len(cache);

    key = rand() % cache->units[0].max_lru_size;
    rc = find_lru_cache(cache, key); 
    //if(rc != key) traceEvent(TRACE_ERROR, "Search found: %u [expected: %u]", rc, key);

    // traceEvent(TRACE_NORMAL, "Cache size [current_size: %u][max_lru_size: %u]", cache->current_size, cache->units[0].max_lru_size);
  }

  rc = find_lru_cache(cache, key); if(rc != key) traceEvent(TRACE_ERROR, "Search found: %u", rc);
  validate_lru_len(cache);
  /* ****************** */

  rc = add_to_lru_cache(cache, key, val); if(rc != 0) traceEvent(TRACE_ERROR, "add_to_lru_cache returned %d", rc);
  validate_lru_len(cache);

  rc = find_lru_cache(cache, key); if(rc != val) traceEvent(TRACE_ERROR, "Search found: %u", rc);
  validate_lru_len(cache);

  traceEvent(TRACE_NORMAL, "Test End");

  val = cache->units[0].max_lru_size;
  free_lru_cache(cache);
  init_lru_cache(cache, val);

  exit(0);
}

/* ************************************ */

int init_lru_cache_unit(struct LruCacheUnit *cache_unit, u_int32_t max_size) {
  traceLRU = readOnlyGlobals.enable_debug;

  if(unlikely(traceLRU)) traceEvent(TRACE_NORMAL, "%s()", __FUNCTION__);

  cache_unit->max_lru_size = max_size, cache_unit->hash_size = 2 * max_size, cache_unit->current_size = 0;
  
  if((cache_unit->hash = (struct LruCacheNode**)calloc(cache_unit->hash_size, sizeof(struct LruCacheNode*))) == NULL) {
    traceEvent(TRACE_ERROR, "Not enough memory?");
    return(-1);
  }
  
  cache_unit->list_head = cache_unit->list_tail = NULL;

  pthread_rwlock_init(&cache_unit->lruLock, NULL);
}

/* ************************************ */

int init_lru_cache(struct LruCache *cache, u_int32_t max_size) {
  int i;

  for(i=0; i<NUM_LRU_CACHE_UNITS; i++)
    init_lru_cache_unit(&cache->units[i], max_size);
}

/* ************************************ */

void free_lru_cache_unit(struct LruCacheUnit *cache_unit) {
  struct LruCacheNode *head = cache_unit->list_head;
  
  if(unlikely(traceLRU)) traceEvent(TRACE_NORMAL, "%s()", __FUNCTION__);

  while(head != NULL) {
    struct LruCacheNode *next = head->lru_list.next;
    
    free(head);
    head = next;
  }

  free(cache_unit->hash);
  pthread_rwlock_destroy(&cache_unit->lruLock);
}

/* ************************************ */

void free_lru_cache(struct LruCache *cache) {
  int i;

  for(i=0; i<NUM_LRU_CACHE_UNITS; i++)
    free_lru_cache_unit(&cache->units[i]);
}

/* ************************************ */

static int delete_oldest_lru_cache_unit(struct LruCacheUnit *cache_unit) {
  struct LruCacheNode *node = cache_unit->list_tail;
  u_int32_t hash_id;
  struct LruCacheNode *head, *prev = NULL;

  if(unlikely(traceLRU)) traceEvent(TRACE_NORMAL, "%s()", __FUNCTION__);

  /* [1] Remove the last list element */
  if(cache_unit->list_tail == NULL) {
    traceEvent(TRACE_ERROR, "Internal error (NULL tail)");
    return(-1);
  } else {
    cache_unit->list_tail = (cache_unit->list_tail)->lru_list.prev;

    if(cache_unit->list_tail == NULL) {
      traceEvent(TRACE_ERROR, "Internal error (NULL tail)");
      return(-1);
    }

    if((cache_unit->list_tail)->lru_list.prev == NULL) {
      traceEvent(TRACE_ERROR, "Internal error (NULL prev)");
      return(-1);
    } else
      ((cache_unit->list_tail)->lru_list.prev)->lru_list.next = NULL;  
  }

  hash_id = node->node.key % cache_unit->hash_size;
  head = cache_unit->hash[hash_id];

  /* [2] Remove the node from the hash */
  while(head != NULL) {
    if(head->node.key == node->node.key) {
      /* Duplicated key found */
      if(prev == NULL) 
	cache_unit->hash[hash_id] = node->hash.next;
      else {
	prev->hash.next = node->hash.next;
      }

      break;
    } else {
      prev = head;
      head = head->hash.next;
    }
  }

  /* [3] Free the memory */
  free(node);
  cache_unit->current_size--;
}

/* ************************************ */

static void delete_node_from_lru_list(struct LruCacheUnit *cache_unit,
				      struct LruCacheNode *node) {
  if(node->lru_list.prev != NULL) {
    (node->lru_list.prev)->lru_list.next = node->lru_list.next;

    if(node->lru_list.next != NULL)
      (node->lru_list.next)->lru_list.prev = node->lru_list.prev;
  } else {
    cache_unit->list_head = node->lru_list.next;
    cache_unit->list_head->lru_list.prev = NULL;
  }
  
  if(cache_unit->list_tail == node) {
    struct LruCacheNode *new_tail = (cache_unit->list_tail)->lru_list.prev;

    if(new_tail == NULL)
      traceEvent(TRACE_ERROR, "Internal error (NULL new_tail)");
    else
      new_tail->lru_list.next = NULL;

    cache_unit->list_tail = new_tail;
  }
}

/* ************************************ */

static void add_node_to_lru_list(struct LruCacheUnit *cache_unit, 
				 struct LruCacheNode *node) {
  if(cache_unit->list_head != NULL) {
    node->lru_list.next = cache_unit->list_head, node->lru_list.prev = NULL;
    (cache_unit->list_head)->lru_list.prev = node;
  } else {
    /*
      The list is empty so our node is going
      to be the oldest one in the LRU 
    */
    
    cache_unit->list_tail = node;
    node->lru_list.next = NULL, node->lru_list.prev = NULL;
  }
  
  cache_unit->list_head = node; /* Add as head */

  if(cache_unit->current_size > cache_unit->max_lru_size)
    delete_oldest_lru_cache_unit(cache_unit); 
}

/* ************************************ */

struct LruCacheNode* allocCacheNode(u_int64_t key, u_int32_t value) {
  struct LruCacheNode *node = (struct LruCacheNode*)calloc(1, sizeof(struct LruCacheNode));
  
  if(unlikely(traceLRU))
    traceEvent(TRACE_NORMAL, "%s(key=%lu, value=%u)", __FUNCTION__, key, value);
  
  if(node == NULL)
    traceEvent(TRACE_ERROR, "Not enough memory?");
  else
    node->node.key = key, node->node.value = value;
  
  return(node);  
}

/* ************************************ */

static int add_to_lru_cache_unit(struct LruCacheUnit *cache_unit, 
				 u_int64_t key, u_int32_t value) {
  u_int32_t hash_id = key % cache_unit->hash_size;
  struct LruCacheNode *node;
  u_int8_t node_already_existing = 0, add_node = 0;
  int rc = 0;

  if(unlikely(traceLRU))
    traceEvent(TRACE_NORMAL, "%s(key=%lu, value=%u)", __FUNCTION__, key, value);

  pthread_rwlock_wrlock(&cache_unit->lruLock);

  /* [1] Add to hash */
  if(cache_unit->hash[hash_id] == NULL) {
    if((node = allocCacheNode(key, value)) == NULL) {
      rc = -1;
      goto ret_add_to_lru_cache;
    }

    cache_unit->hash[hash_id] = node;
    cache_unit->current_size++;
    add_node_to_lru_list(cache_unit, node);
  } else {
    /* Check if the element exists */
    struct LruCacheNode *head = cache_unit->hash[hash_id];
      
    while(head != NULL) {
      if(head->node.key == key) {
	/* Duplicated key found */
	node = head;
	node->node.value = value; /* Overwrite old value */
	node_already_existing = 1;
	break;
      } else
	head = head->hash.next;
    }

    if(!node_already_existing) {
      if((node = allocCacheNode(key, value)) == NULL) {
	rc = -2;
	goto ret_add_to_lru_cache;
      }

      node->hash.next = cache_unit->hash[hash_id];
      cache_unit->hash[hash_id] = node;
      cache_unit->current_size++;
      add_node_to_lru_list(cache_unit, node);
    }
  }

 ret_add_to_lru_cache:
  pthread_rwlock_unlock(&cache_unit->lruLock);
  return(rc);
}

/* ************************************ */

int add_to_lru_cache(struct LruCache *cache, u_int64_t key, u_int32_t value) {
  return(add_to_lru_cache_unit(&cache->units[key % NUM_LRU_CACHE_UNITS], key, value));
}

/* ************************************ */

static u_int32_t find_lru_cache_unit(struct LruCacheUnit *cache_unit, u_int64_t key) {
  u_int32_t hash_id = key % cache_unit->hash_size;
  struct LruCacheNode *head = cache_unit->hash[hash_id];
  u_int32_t ret_val = NDPI_PROTOCOL_UNKNOWN;

  if(unlikely(traceLRU)) 
    traceEvent(TRACE_NORMAL, "%s(%lu)", __FUNCTION__, key);
    
  pthread_rwlock_rdlock(&cache_unit->lruLock);

  while(head != NULL) {
    if(head->node.key == key) {
      ret_val = head->node.value;
      break;
    } else {
      head = head->hash.next;
    }
  }

#if 0
  /* 
     In theory we should do this, but in practice it is not
     necessary as when a flow expires, we re-add it to the cache
     and thus in case of trashing it will be restored
     by that add anyway
  */

  if((head != NULL /* We have found the value */) 
     && (cache_unit->list_head != head) /* We're not already the head */) {
    /* Remove element from LRU */
    delete_node_from_lru_list(cache_unit, head);
    
    /* Add it as head of the LRU */
    add_node_to_lru_list(cache_unit, head);
  }
#endif

  pthread_rwlock_unlock(&cache_unit->lruLock);

  return(ret_val);
}

/* ************************************ */

u_int32_t find_lru_cache(struct LruCache *cache, u_int64_t key) {
  return(find_lru_cache_unit(&cache->units[key % NUM_LRU_CACHE_UNITS], key));
}
